#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#include <pgm/pgm.h>

#include "ppport.h"


typedef struct pgm_transport_info_t pgm_transport_info_t;

static
pgm_transport_info_t*
tinfo_sv2c (
	SV*				tinfo_ref,
	pgm_transport_info_t**		tinfo_ptr
	)
{
	SV **recv_svp, **send_svp, **svp;
	gsize recv_addrs_len = 0, send_addrs_len = 0;
	pgm_transport_info_t *ti = NULL;
	int i;

	if (tinfo_ref && SvROK (tinfo_ref) && tinfo_ptr)
	{
/* recv address list */
		recv_svp = hv_fetch ((HV*)SvRV (tinfo_ref), "recv_addrs", strlen("recv_addrs"), 0);
		if (recv_svp && (SvTYPE (SvRV (*recv_svp)) == SVt_PVAV))
			recv_addrs_len = av_len ((AV*)SvRV (*recv_svp)) + 1;
/* send address list */
		send_svp = hv_fetch ((HV*)SvRV (tinfo_ref), "send_addrs", strlen("send_addrs"), 0);
		if (send_svp && (SvTYPE (SvRV (*send_svp)) == SVt_PVAV))
			send_addrs_len = av_len ((AV*)SvRV (*send_svp)) + 1;
	
		ti = g_malloc0 (sizeof(struct pgm_transport_info_t) +
				(recv_addrs_len + send_addrs_len) * sizeof(struct group_source_req));
		svp = hv_fetch ((HV*)SvRV (tinfo_ref), "gsi", strlen("gsi"), 0);
		if (svp)
			memcpy (&ti->ti_gsi, SvPV_nolen (*svp), SvCUR (*svp));
		svp = hv_fetch ((HV*)SvRV (tinfo_ref), "udp_encap_ucast_port", strlen("udp_encap_ucast_port"), 0);
		ti->ti_udp_encap_ucast_port = (svp && SvIOK (*svp)) ? SvUV (*svp) : 0;
		svp = hv_fetch ((HV*)SvRV (tinfo_ref), "udp_encap_mcast_port", strlen("udp_encap_mcast_port"), 0);
		ti->ti_udp_encap_mcast_port = (svp && SvIOK (*svp)) ? SvUV (*svp) : 0;
		svp = hv_fetch ((HV*)SvRV (tinfo_ref), "sport", strlen("sport"), 0);
		ti->ti_sport = (svp && SvIOK (*svp)) ? SvUV (*svp) : 0;
		svp = hv_fetch ((HV*)SvRV (tinfo_ref), "dport", strlen("dport"), 0);
		ti->ti_dport = (svp && SvIOK (*svp)) ? SvUV (*svp) : 0;
		if (recv_addrs_len) {
			AV *gsr_array = (AV*)SvRV (*recv_svp);
			ti->ti_recv_addrs = (gpointer)((guint8*)ti + sizeof(struct pgm_transport_info_t));
			for (i = 0; i < recv_addrs_len; i++) {
				SV **gsr = av_fetch (gsr_array, i, 0);
				svp = hv_fetch ((HV*)SvRV (*gsr), "interface", strlen("interface"), 0);
				ti->ti_recv_addrs[i].gsr_interface = (svp && SvIOK (*svp)) ? SvUV (*svp) : 0;
				svp = hv_fetch ((HV*)SvRV (*gsr), "group", strlen("group"), 0);
				if (svp)
					memcpy (&ti->ti_recv_addrs[i].gsr_group, SvPV_nolen (*svp), SvCUR (*svp));
				svp = hv_fetch ((HV*)SvRV (*gsr), "source", strlen("source"), 0);
				if (svp)
					memcpy (&ti->ti_recv_addrs[i].gsr_source, SvPV_nolen (*svp), SvCUR (*svp));
			}
			ti->ti_recv_addrs_len = recv_addrs_len;
		}
		if (send_addrs_len) {
			AV *gsr_array = (AV*)SvRV (*send_svp);
			ti->ti_send_addrs = (gpointer)((guint8*)ti->ti_recv_addrs + recv_addrs_len * sizeof(struct group_source_req));
			for (i = 0; i < send_addrs_len; i++) {
				SV **gsr = av_fetch (gsr_array, i, 0);
				svp = hv_fetch ((HV*)SvRV (*gsr), "interface", strlen("interface"), 0);
				ti->ti_send_addrs[i].gsr_interface = (svp && SvIOK (*svp)) ? SvUV (*svp) : 0;
				svp = hv_fetch ((HV*)SvRV (*gsr), "group", strlen("group"), 0);
				if (svp)
					memcpy (&ti->ti_send_addrs[i].gsr_group, SvPV_nolen (*svp), SvCUR (*svp));
				svp = hv_fetch ((HV*)SvRV (*gsr), "source", strlen("source"), 0);
				if (svp)
					memcpy (&ti->ti_send_addrs[i].gsr_source, SvPV_nolen (*svp), SvCUR (*svp));
			}
			ti->ti_send_addrs_len = send_addrs_len;
		}

		*tinfo_ptr = ti;
	}
	return ti;
}

static
SV*
tinfo_c2sv (
	pgm_transport_info_t*		tinfo_ptr
	)
{
	HV *gsr, *tinfo = newHV();
	AV *addrs;
	int i;

/* recv address list */
	addrs = newAV();
	for (i = 0; i < tinfo_ptr->ti_recv_addrs_len; i++) {
		gsr = newHV();
		hv_store (gsr, "interface", strlen("interface"), newSVuv (tinfo_ptr->ti_recv_addrs[i].gsr_interface), 0);
		hv_store (gsr, "group", strlen ("group"), newSVpvn ((const char*)&tinfo_ptr->ti_recv_addrs[i].gsr_group, sizeof(struct sockaddr_storage)), 0);
		hv_store (gsr, "source", strlen ("source"), newSVpvn ((const char*)&tinfo_ptr->ti_recv_addrs[i].gsr_source, sizeof(struct sockaddr_storage)), 0);
		av_store (addrs, i, newRV_noinc ((SV*)gsr));
	}
	hv_store (tinfo, "recv_addrs", strlen ("recv_addrs"), newRV_noinc ((SV*)addrs), 0);
/* send address list */
	addrs = newAV();
	for (i = 0; i < tinfo_ptr->ti_send_addrs_len; i++) {
		gsr = newHV();
		hv_store (gsr, "interface", strlen("interface"), newSVuv (tinfo_ptr->ti_send_addrs[i].gsr_interface), 0);
		hv_store (gsr, "group", strlen ("group"), newSVpvn ((const char*)&tinfo_ptr->ti_send_addrs[i].gsr_group, sizeof(struct sockaddr_storage)), 0);
		hv_store (gsr, "source", strlen ("source"), newSVpvn ((const char*)&tinfo_ptr->ti_send_addrs[i].gsr_source, sizeof(struct sockaddr_storage)), 0);
		av_store (addrs, i, newRV_noinc ((SV*)gsr));
	}
	hv_store (tinfo, "send_addrs", strlen ("send_addrs"), newRV_noinc ((SV*)addrs), 0);

	hv_store (tinfo, "sport", strlen ("sport"), newSVuv (tinfo_ptr->ti_sport), 0);
	hv_store (tinfo, "dport", strlen ("dport"), newSVuv (tinfo_ptr->ti_dport), 0);

	return (SV *)tinfo;
}


MODULE = OpenPgm		PACKAGE = OpenPgm		PREFIX = pgm_

int
pgm_init ()
	CODE:
		RETVAL = !pgm_init ();
	OUTPUT:
		RETVAL

int
pgm_shutdown ()
	CODE:
		RETVAL = !pgm_shutdown ();
	OUTPUT:
		RETVAL

MODULE = OpenPgm		PACKAGE = OpenPgm::Interface	PREFIX = pgm_if_

void
pgm_if_get_transport_info (network, hints_ref=NULL)
		char* network
		SV* hints_ref = (SvOK ($arg) ? $arg : NULL);
	PREINIT:
		GError* err = NULL;
		struct pgm_transport_info_t *hints_ptr = NULL;
		struct pgm_transport_info_t *tinfo_ptr = NULL;
		gboolean res;
	PPCODE:
		if (hints_ref && (NULL == tinfo_sv2c (hints_ref, &hints_ptr)))
			XSRETURN_UNDEF;
		res = pgm_if_get_transport_info (network, hints_ptr, &tinfo_ptr, &err);
		pgm_if_free_transport_info (hints_ptr);
		if (err)
			gperl_croak_gerror (NULL, err);
		XPUSHs (sv_2mortal (newRV_noinc (tinfo_c2sv (tinfo_ptr))));

MODULE = OpenPgm		PACKAGE = OpenPgm::Gsi		PREFIX = pgm_gsi_

void
pgm_gsi_create_from_hostname ()
	PREINIT:
		pgm_gsi_t gsi;
		GError* err = NULL;
		gboolean res;
	PPCODE:
		res = pgm_gsi_create_from_hostname (&gsi, &err);
		if (err)
			gperl_croak_gerror (NULL, err);
		XPUSHs (sv_2mortal (newSVpv ((const char*)&gsi, sizeof(gsi))));

MODULE = OpenPgm		PACKAGE = OpenPgm::Transport	PREFIX = pgm_transport_

SV*
pgm_transport_create (tinfo_ref)
		SV*		tinfo_ref
	PREINIT:
		GError* err = NULL;
		struct pgm_transport_info_t *tinfo_ptr = NULL;
		pgm_transport_t* transport;
		int res;
	PPCODE:
		if (NULL == tinfo_sv2c (tinfo_ref, &tinfo_ptr))
			XSRETURN_UNDEF;
		res = pgm_transport_create (&transport, tinfo_ptr, &err);
		pgm_if_free_transport_info (tinfo_ptr);
		if (err)
			gperl_croak_gerror (NULL, err);

		HV* stash = gv_stashpv("OpenPgm::Transport", GV_ADDWARN);
		SV* object;
		HV* options = newHV();

		hv_store (options, "transport", strlen("transport"), newSVpvn ((const char*)&transport, sizeof(pgm_transport_t*)), 0);
		object = newRV_noinc ((SV*)options);
		sv_bless (object, stash);
		XPUSHs(sv_2mortal (object));

gboolean
pgm_transport_bind (class)
		HV* class
	PREINIT:
		pgm_transport_t* transport = NULL;
		GError* err = NULL;
		SV** svp;
	CODE:
		if ((svp = hv_fetch (class, "transport", strlen("transport"), FALSE)))
			transport = *(pgm_transport_t**)SvPV_nolen (*svp);
		RETVAL = pgm_transport_bind (transport, &err);
		if (err)
			gperl_croak_gerror (NULL, err);
	OUTPUT:
		RETVAL

gboolean
pgm_transport_destroy (class, flush)
		HV* class
		const gboolean flush
	PREINIT:
		pgm_transport_t* transport;
		SV** svp;
	CODE:
		if ((svp = hv_fetch (class, "transport", strlen("transport"), FALSE)))
			transport = *(pgm_transport_t**)SvPV_nolen (*svp);
		RETVAL = pgm_transport_destroy (transport, flush);
	OUTPUT:
		RETVAL

gboolean
pgm_transport_set_send_only (class, send_only)
		HV* class
		const gboolean send_only
	PREINIT:
		pgm_transport_t* transport = NULL;
		SV** svp;
	CODE:
		if ((svp = hv_fetch (class, "transport", strlen("transport"), FALSE)))
			transport = *(pgm_transport_t**)SvPV_nolen (*svp);
		RETVAL = pgm_transport_set_send_only (transport, send_only);
	OUTPUT:
		RETVAL

gboolean
pgm_transport_set_max_tpdu (class, max_tpdu)
		HV* class
		const guint16 max_tpdu
	PREINIT:
		pgm_transport_t* transport = NULL;
		SV** svp;
	CODE:
		if ((svp = hv_fetch (class, "transport", strlen("transport"), FALSE)))
			transport = *(pgm_transport_t**)SvPV_nolen (*svp);
		RETVAL = pgm_transport_set_max_tpdu (transport, max_tpdu);
	OUTPUT:
		RETVAL

gboolean
pgm_transport_set_txw_sqns (class, sqns)
		HV* class
		const guint sqns
	PREINIT:
		pgm_transport_t* transport = NULL;
		SV** svp;
	CODE:
		if ((svp = hv_fetch (class, "transport", strlen("transport"), FALSE)))
			transport = *(pgm_transport_t**)SvPV_nolen (*svp);
		RETVAL = pgm_transport_set_txw_sqns (transport, sqns);
	OUTPUT:
		RETVAL

gboolean
pgm_transport_set_txw_max_rte (class, rte)
		HV* class
		const guint rte
	PREINIT:
		pgm_transport_t* transport = NULL;
		SV** svp;
	CODE:
		if ((svp = hv_fetch (class, "transport", strlen("transport"), FALSE)))
			transport = *(pgm_transport_t**)SvPV_nolen (*svp);
		RETVAL = pgm_transport_set_txw_max_rte (transport, rte);
	OUTPUT:
		RETVAL

gboolean
pgm_transport_set_multicast_loop (class, use_multicast_loop)
		HV* class
		const gboolean use_multicast_loop
	PREINIT:
		pgm_transport_t* transport = NULL;
		SV** svp;
	CODE:
		if ((svp = hv_fetch (class, "transport", strlen("transport"), FALSE)))
			transport = *(pgm_transport_t**)SvPV_nolen (*svp);
		RETVAL = pgm_transport_set_multicast_loop (transport, use_multicast_loop);
	OUTPUT:
		RETVAL

gboolean
pgm_transport_set_hops (class, hops)
		HV* class
		const guint hops
	PREINIT:
		pgm_transport_t* transport = NULL;
		SV** svp;
	CODE:
		if ((svp = hv_fetch (class, "transport", strlen("transport"), FALSE)))
			transport = *(pgm_transport_t**)SvPV_nolen (*svp);
		RETVAL = pgm_transport_set_hops (transport, hops);
	OUTPUT:
		RETVAL

gboolean
pgm_transport_set_ambient_spm (class, interval)
		HV* class
		const guint interval
	PREINIT:
		pgm_transport_t* transport = NULL;
		SV** svp;
	CODE:
		if ((svp = hv_fetch (class, "transport", strlen("transport"), FALSE)))
			transport = *(pgm_transport_t**)SvPV_nolen (*svp);
		RETVAL = pgm_transport_set_ambient_spm (transport, interval);
	OUTPUT:
		RETVAL

gboolean
pgm_transport_set_heartbeat_spm (class, intervals_ref)
		HV* class
		SV* intervals_ref
	INIT:
		pgm_transport_t* transport = NULL;
		SV** svp;
		guint* intervals;
		I32 numintervals = 0;
		int i;
	CODE:
		if ((svp = hv_fetch (class, "transport", strlen("transport"), FALSE)))
			transport = *(pgm_transport_t**)SvPV_nolen (*svp);
		if (! SvROK(intervals_ref)
		    || (SvTYPE(SvRV(intervals_ref)) != SVt_PVAV)
                    || ((numintervals = av_len((AV *)SvRV(intervals_ref))) < 0))
		{
			XSRETURN_UNDEF;
		}
		intervals = g_new (guint, numintervals);
		for (i = 0; i < numintervals; i++) {
			SV **interval_svp = av_fetch ((AV*)SvRV (intervals_ref), i, 0);
			intervals[i] = (SvOK (*interval_svp) ? SvUV (*interval_svp) : 0);
		}
		RETVAL = pgm_transport_set_heartbeat_spm (transport, intervals, numintervals);
	OUTPUT:
		RETVAL

MODULE = OpenPgm		PACKAGE = OpenPgm::Transport	PREFIX = pgm_

PGMIOStatus
pgm_send (class, apdu, apdu_length, bytes_written_ref=NULL)
		HV* class
		SV* apdu
		const gsize apdu_length
		SV* bytes_written_ref = (SvOK ($arg) ? $arg : NULL);
	PREINIT:
		pgm_transport_t* transport = NULL;
		SV** svp;
		gsize bytes_written = 0;
	CODE:
		if ((svp = hv_fetch (class, "transport", strlen("transport"), FALSE)))
			transport = *(pgm_transport_t**)SvPV_nolen (*svp);
		RETVAL = pgm_send (transport, SvPV(apdu, PL_na), apdu_length, &bytes_written);
	OUTPUT:
		RETVAL

